(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: Cull.cpp
//
// Desc: Shows a technique for culling objects whose bounding boxes are
//       outside the view frustum.  This technique is described at:
//       http://www.cs.unc.edu/~hoff/research/vfculler/viewcull.html
//
//       Note: This code uses the D3D Framework helper library.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit CullUnit;

interface

uses
  Windows, Messages, SysUtils, Math, CommDlg,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFont, D3DUtil, DXUtil;

type
  //---------------------------------------------------------------------------
  // Name: enum CULLSTATE
  // Desc: Represents the result of the culling calculation on an object.
  //---------------------------------------------------------------------------
  TCullState = (
    CS_UNKNOWN,      // cull state not yet computed
    CS_INSIDE,       // object bounding box is at least partly inside the frustum
    CS_OUTSIDE,      // object bounding box is outside the frustum
    CS_INSIDE_SLOW,  // OBB is inside frustum, but it took extensive testing to determine this
    CS_OUTSIDE_SLOW  // OBB is outside frustum, but it took extensive testing to determine this
  );


  //---------------------------------------------------------------------------
  // Name: struct CULLINFO
  // Desc: Stores information that will be used when culling objects.  It needs
  //       to be recomputed whenever the view matrix or projection matrix changes.
  //---------------------------------------------------------------------------
  PCullInfo = ^TCullInfo;
  TCullInfo = record
    vecFrustum: array[0..7] of TD3DXVector3;    // corners of the view frustum
    planeFrustum: array[0..5] of TD3DXPlane;    // planes of the view frustum
  end;

  PD3DXVector3Array = ^TD3DXVector3Array;
  TD3DXVector3Array = array[0..MaxInt div SizeOf(TD3DXVector3) - 1] of TD3DXVector3;

  PD3DXPlaneArray = ^TD3DXPlaneArray;
  TD3DXPlaneArray = array[0..MaxInt div SizeOf(TD3DXPlane) - 1] of TD3DXPlane;


// Prototypes for the culling functions
procedure UpdateCullInfo(var pCullInfo: TCullInfo; const pMatView, pMatProj: TD3DXMatrixA16);
function CullObject(const pCullInfo: TCullInfo; pVecBounds: PD3DXVector3Array; pPlaneBounds: PD3DXPlaneArray): TCullState;
function EdgeIntersectsFace(pEdges, pFaces: PD3DXVector3Array; const pPlane: TD3DXPlane): BOOL;


type
  //---------------------------------------------------------------------------
  // Name: struct PLANEVERTEX
  // Desc: Custom vertex type used for drawing the frustum planes
  //---------------------------------------------------------------------------
  PPlaneVertex = ^TPlaneVertex;
  TPlaneVertex = packed record
    p: TD3DXVector3;
    color: DWORD;
  end;
  PPlaneVertexArray = ^TPlaneVertexArray;
  TPlaneVertexArray = array[0..MaxInt div SizeOf(TPlaneVertex) - 1] of TPlaneVertex;

const
  TPlaneVertex_FVF = (D3DFVF_XYZ or D3DFVF_DIFFUSE);


type
  //---------------------------------------------------------------------------
  // Name: class CCullableThing
  // Desc: A cullable object
  //---------------------------------------------------------------------------
  CCullableThing = class
  public
    m_pos:   TD3DXVector3;    // origin of object
    m_fRotX: Single;          // rotation of object around X axis
    m_fRotY: Single;          // rotation of object around Y axis
    m_mat:   TD3DXMatrix;     // object's local-to-world transformation
    m_vecBoundsLocal: array[0..7] of TD3DXVector3; // bounding box coordinates (in local coord space)
    m_vecBoundsWorld: array[0..7] of TD3DXVector3; // bounding box coordinates (in world coord space)
    m_planeBoundsWorld: array[0..5] of TD3DXPlane; // bounding box planes (in world coord space)
    m_cullstate: TCullState; // whether object is in the view frustum

  public
    procedure Init;
    procedure UpdateMatrix;
  end;

  //---------------------------------------------------------------------------
  // Name: struct Camera
  // Desc:
  //---------------------------------------------------------------------------
  TCamera = record
    m_vPosition:      TD3DXVector3;
    m_vVelocity:      TD3DXVector3;
    m_fYaw:           Single;
    m_fYawVelocity:   Single;
    m_fPitch:         Single;
    m_fPitchVelocity: Single;
    m_matView:        TD3DXMatrix;
    m_matOrientation: TD3DXMatrix;
  end;




  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:       CD3DFont;      // Font for drawing text
    m_pFontSmall:  CD3DFont;

    m_pMeshTeapot: ID3DXMesh;     // Mesh of thing to be cull-tested
    m_pMeshBox:    ID3DXMesh;     // Mesh to visualize bounding box
    m_matBox:      TD3DXMatrix;   // Matrix that places bounding box correctly on teapot

    m_pPlaneVB: array[0..5] of IDirect3DVertexBuffer9; // VBs to visualize the view frustum

    // The things to render and cull
    m_CullableThingArray: array[0..49] of CCullableThing;
    m_dwNumCullableThings:      DWORD;

    // Variables for determining the view
    m_bKey:             array[#0..#255] of Boolean;
    m_bLeftActive:      Boolean;  // left vs right view currently active
    m_CameraLeft:       TCamera;
    m_CameraRight:      TCamera;
    m_matProjLeft:      TD3DXMatrix;
    m_matProjRight:     TD3DXMatrix;

    m_mtrlOutside:      TD3DMaterial9;
    m_mtrlInside:       TD3DMaterial9;
    m_mtrlOutsideSlow:  TD3DMaterial9;
    m_mtrlInsideSlow:   TD3DMaterial9;
    m_mtrlWhite:        TD3DMaterial9;

    m_cullinfo:         TCullInfo;
    m_bShowHelp:        Boolean;

    procedure UpdateCamera(var pCamera: TCamera);
    function UpdatePlaneVBs: HResult;
    procedure CullObjects;
  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function RenderScene(bRenderPlanes: BOOL): HResult;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

implementation

procedure CCullableThing.Init;
begin
  // Pick a random position and orientation
  m_pos := D3DXVector3(Random*50 - 25, // X is in (-25.0, 25.0)
                       Random*50 - 25, // Y is in (-25.0, 25.0)
                       Random*25);     // Z is in (  0.0, 25.0)
  m_fRotX := D3DXToRadian(Random * 360);
  m_fRotY := D3DXToRadian(Random * 360);
  UpdateMatrix;
  m_cullstate := CS_UNKNOWN;
end;

procedure CCullableThing.UpdateMatrix;
var
  matRotX, matRotY, matTrans: TD3DXMatrix;
  i: Integer;
  m: TD3DXmatrix;
begin
  // Recompute m_mat, m_vecBoundsWorld, and m_planeBoundsWorld
  // when the thing's position, orientation, or bounding box has changed
  D3DXMatrixRotationX(matRotX, m_fRotX);
  D3DXMatrixRotationY(matRotY, m_fRotY);
  D3DXMatrixTranslation(matTrans, m_pos.x, m_pos.y, m_pos.z);
  // m_mat := matRotX * matRotY * matTrans;
  D3DXMatrixMultiply(m, matRotY, matTrans);
  D3DXMatrixMultiply(m_mat, matRotX, m);

  // Transform bounding box coords from local space to world space
  for i:= 0 to 7 do
    D3DXVec3TransformCoord(m_vecBoundsWorld[i], m_vecBoundsLocal[i], m_mat);

  // Determine planes of the bounding box
  D3DXPlaneFromPoints(m_planeBoundsWorld[0], m_vecBoundsWorld[0],
    m_vecBoundsWorld[1], m_vecBoundsWorld[2]); // Near
  D3DXPlaneFromPoints(m_planeBoundsWorld[1], m_vecBoundsWorld[6],
    m_vecBoundsWorld[7], m_vecBoundsWorld[5]); // Far
  D3DXPlaneFromPoints(m_planeBoundsWorld[2], m_vecBoundsWorld[2],
    m_vecBoundsWorld[6], m_vecBoundsWorld[4]); // Left
  D3DXPlaneFromPoints(m_planeBoundsWorld[3], m_vecBoundsWorld[7],
    m_vecBoundsWorld[3], m_vecBoundsWorld[5]); // Right
  D3DXPlaneFromPoints(m_planeBoundsWorld[4], m_vecBoundsWorld[2],
    m_vecBoundsWorld[3], m_vecBoundsWorld[6]); // Top
  D3DXPlaneFromPoints(m_planeBoundsWorld[5], m_vecBoundsWorld[1],
    m_vecBoundsWorld[0], m_vecBoundsWorld[4]); // Bottom
end;


{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle := 'Cull: Culling nonvisible objects';

  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_dwCreationWidth     := 600;
  m_dwCreationHeight    := 300;
  m_bShowCursorWhenFullscreen := True;

  m_pFont               := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall          := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);
  m_pMeshTeapot         := nil;
  m_pMeshBox            := nil;
  m_dwNumCullableThings := 0;

  ZeroMemory(@m_pPlaneVB, SizeOf(m_pPlaneVB));
  ZeroMemory(@m_bKey, 256);
  ZeroMemory(@m_CameraLeft, SizeOf(m_CameraLeft));
  ZeroMemory(@m_CameraRight, SizeOf(m_CameraRight));

  m_bLeftActive := True;
  m_bShowHelp := False;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
var
  i: Integer;
begin
  // Initialize cullable things
  for i := 0 to 49 do
  begin
    m_CullableThingArray[m_dwNumCullableThings]:= CCullableThing.Create;
    m_CullableThingArray[m_dwNumCullableThings].Init;
    Inc(m_dwNumCullableThings);
  end;

  UpdateCamera(m_CameraLeft);
  UpdateCamera(m_CameraRight);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  bNeedToCull: BOOL;
  pCullableThing: CCullableThing;
  iThing: DWORD;
begin
  bNeedToCull := False;

  // Handle object rotations
  if (m_bKey['Y'] or m_bKey['U'] or m_bKey['H'] or m_bKey['J']) then
  begin
    for iThing:= 0 to m_dwNumCullableThings - 1 do
    begin
      pCullableThing := m_CullableThingArray[iThing];
      if m_bKey['Y'] then pCullableThing.m_fRotY:= pCullableThing.m_fRotY + m_fElapsedTime
      else if m_bKey['U'] then pCullableThing.m_fRotY:= pCullableThing.m_fRotY - m_fElapsedTime;

      if m_bKey['H'] then pCullableThing.m_fRotX:= pCullableThing.m_fRotX + m_fElapsedTime
      else if m_bKey['J'] then pCullableThing.m_fRotX:= pCullableThing.m_fRotX - m_fElapsedTime;

      pCullableThing.UpdateMatrix;
    end;
    bNeedToCull := True;
  end;

  // Handle camera motion
  if (m_bLeftActive) then
  begin
    UpdateCamera(m_CameraLeft);
  end else
  begin
    UpdateCamera(m_CameraRight);
    UpdateCullInfo(m_cullinfo, m_CameraRight.m_matView, m_matProjRight);
    bNeedToCull := True;
    UpdatePlaneVBs;
  end;

  // Re-determine cull state of all objects if necessary
  if bNeedToCull then CullObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
const
  // The hot clear color is used to indicate the viewport
  // whose camera is currently being controlled by the keyboard
  dwHotClearColor  = TD3DColor($ff0000ff);
  dwColdClearColor = TD3DColor($ff000080);

  IsHotColor: array[False..True] of TD3DColor = (dwColdClearColor, dwHotClearColor);
var
  vp: TD3DViewport9;
begin
  //
  // Draw left viewport
  //
  vp.X := 0;
  vp.Y := 0;
  vp.Width  := m_d3dsdBackBuffer.Width div 2;
  vp.Height := m_d3dsdBackBuffer.Height;
  vp.MinZ := 0.0;
  vp.MaxZ := 1.0;
  m_pd3dDevice.SetViewport(vp);

  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
    IsHotColor[m_bLeftActive], 1.0, 0);

  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_CameraLeft.m_matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProjLeft);

  // Draw contents of left viewport
  RenderScene(True); // TRUE means render frustum planes

  //
  // Draw right viewport
  //
  vp.X := m_d3dsdBackBuffer.Width div 2;
  vp.Width := m_d3dsdBackBuffer.Width div 2;
  m_pd3dDevice.SetViewport(vp);

  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
    IsHotColor[not m_bLeftActive], 1.0, 0);

  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_CameraRight.m_matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProjRight);

  // Draw contents of right viewport
  RenderScene(False); // FALSE means don't render frustum planes

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc: 
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderScene(bRenderPlanes: BOOL): HResult;
var
  pCullableThing: CCullableThing;
  iThing: DWORD;
  mat: TD3DXMatrix;
  iPlane: Integer;
begin
  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Render each object
    for iThing:= 0 to m_dwNumCullableThings - 1 do
    begin
      pCullableThing:= m_CullableThingArray[iThing];

      // Normally, if the cullstate is CS_OUTSIDE or CS_OUTSIDE_SLOW,
      // there is no need to ask D3D to render the object since
      // it's outside the view frustum.  Since this app is meant to
      // visualize the culling process, all objects are passed to D3D,
      // and different colors are used to show their cullstates.

      case pCullableThing.m_cullstate of
        CS_UNKNOWN:      m_pd3dDevice.SetMaterial(m_mtrlWhite);
        CS_OUTSIDE:      m_pd3dDevice.SetMaterial(m_mtrlOutside);
        CS_INSIDE:       m_pd3dDevice.SetMaterial(m_mtrlInside);
        CS_OUTSIDE_SLOW: m_pd3dDevice.SetMaterial(m_mtrlOutsideSlow);
        CS_INSIDE_SLOW:  m_pd3dDevice.SetMaterial(m_mtrlInsideSlow);
      end;
      m_pd3dDevice.SetTransform(D3DTS_WORLD, pCullableThing.m_mat);
      m_pMeshTeapot.DrawSubset(0);
      // mat = m_matBox * pCullableThing->m_mat;
      D3DXMatrixMultiply(mat, m_matBox, pCullableThing.m_mat);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, mat);
      m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
      m_pMeshBox.DrawSubset(0);
      m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
    end;

    if bRenderPlanes then
    begin
      // Render frustum planes
      m_pd3dDevice.SetMaterial(m_mtrlWhite);
      D3DXMatrixIdentity(mat);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, mat);
      m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
      m_pd3dDevice.SetFVF(TPlaneVertex_FVF);
      m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
      for iPlane:= 0 to 5 do
      begin
        m_pd3dDevice.SetStreamSource(0, m_pPlaneVB[iPlane], 0, SizeOf(TPlaneVertex));
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
      end;
      m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
      m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    end;

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
        
    // Show help
    if m_bShowHelp then
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,200,100,100),
                             'Press F1 to hide help'#10 +
                             'The right viewport is the frustum being culled against.'#10 +
                             'The left viewport has its own camera that lets you view'#10 +
                             'the scene and frustum from different angles.'#10 +
                             'Click the viewport whose camera you want to control.'#10 +
                             'Keyboard controls:');
      m_pFontSmall.DrawText(20, 140, D3DCOLOR_ARGB(255,200,100,100),
                             'Move'#10 +
                             'Turn'#10 +
                             'Spin objects'#10 +
                             'Snap left view to viewport'#10 +
                             'Snap right view to origin'#10);
      m_pFontSmall.DrawText(210, 140, D3DCOLOR_ARGB(255,200,100,100),
                             'W, S, Arrow keys'#10 +
                             'Q, E, A, Z'#10 +
                             'Y, U, H, J'#10 +
                             'N'#10 +
                             'M');
    end else
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,200,100,100),
                            'Press F1 for help');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: UpdateCamera()
// Desc:
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.UpdateCamera(var pCamera: TCamera);
var
  fElapsedTime: Single;
  fSpeed, fAngularSpeed: Single;
  vT: TD3DXVector3;
  qR: TD3DXQuaternion;
begin
  if (m_fElapsedTime > 0.0) then
    fElapsedTime := m_fElapsedTime
  else
    fElapsedTime := 0.05;

  fSpeed        := 5.0*fElapsedTime;
  fAngularSpeed := 2.0*fElapsedTime;

  // De-accelerate the camera movement (for smooth motion)
  D3DXVec3Scale(pCamera.m_vVelocity, pCamera.m_vVelocity, 0.75);
  pCamera.m_fYawVelocity   := pCamera.m_fYawVelocity   * 0.75;
  pCamera.m_fPitchVelocity := pCamera.m_fPitchVelocity * 0.75;

  // Process keyboard input
  if m_bKey[Char(VK_RIGHT)] then pCamera.m_vVelocity.x    := pCamera.m_vVelocity.x    + fSpeed; // Slide Right
  if m_bKey[Char(VK_LEFT)]  then pCamera.m_vVelocity.x    := pCamera.m_vVelocity.x    - fSpeed; // Slide Left
  if m_bKey[Char(VK_UP)]    then pCamera.m_vVelocity.y    := pCamera.m_vVelocity.y    + fSpeed; // Slide Up
  if m_bKey[Char(VK_DOWN)]  then pCamera.m_vVelocity.y    := pCamera.m_vVelocity.y    - fSpeed; // Slide Down
  if m_bKey['W']            then pCamera.m_vVelocity.z    := pCamera.m_vVelocity.z    + fSpeed; // Move Forward
  if m_bKey['S']            then pCamera.m_vVelocity.z    := pCamera.m_vVelocity.z    - fSpeed; // Move Backward
  if m_bKey['E']            then pCamera.m_fYawVelocity   := pCamera.m_fYawVelocity   + fSpeed; // Turn Right
  if m_bKey['Q']            then pCamera.m_fYawVelocity   := pCamera.m_fYawVelocity   - fSpeed; // Turn Left
  if m_bKey['Z']            then pCamera.m_fPitchVelocity := pCamera.m_fPitchVelocity + fSpeed; // Turn Down
  if m_bKey['A']            then pCamera.m_fPitchVelocity := pCamera.m_fPitchVelocity - fSpeed; // Turn Up

  // Update the position vector
  D3DXVec3Scale(vT, pCamera.m_vVelocity, fSpeed);
  D3DXVec3TransformNormal(vT, vT, pCamera.m_matOrientation);
  D3DXVec3Add(pCamera.m_vPosition, pCamera.m_vPosition, vT);

  // Update the yaw-pitch-rotation vector
  pCamera.m_fYaw   := pCamera.m_fYaw   + fAngularSpeed * pCamera.m_fYawVelocity;
  pCamera.m_fPitch := pCamera.m_fPitch + fAngularSpeed * pCamera.m_fPitchVelocity;
  if (pCamera.m_fPitch < -D3DX_PI/2) then pCamera.m_fPitch := -D3DX_PI/2;
  if (pCamera.m_fPitch >  D3DX_PI/2) then pCamera.m_fPitch :=  D3DX_PI/2;

  // Set the view matrix
  D3DXQuaternionRotationYawPitchRoll(qR, pCamera.m_fYaw, pCamera.m_fPitch, 0.0);
  D3DXMatrixAffineTransformation(pCamera.m_matOrientation, 1.25, nil, @qR, @pCamera.m_vPosition);
  D3DXMatrixInverse(pCamera.m_matView, nil, pCamera.m_matOrientation);
end;

//-----------------------------------------------------------------------------
// Name: UpdatePlaneVBs()
// Desc: Update the vertex buffers to match the view frustum.  Slightly
//       different colors are used for each plane to make them easier to
//       distinguish from each other.
//-----------------------------------------------------------------------------
function CMyD3DApplication.UpdatePlaneVBs: HResult;
var
  iPlane: Integer;
  v: PPlaneVertexArray;
begin
  for iPlane:= 0 to 5 do
  begin
    Result:= m_pPlaneVB[iPlane].Lock(0, 0, Pointer(v), 0);
    if FAILED(Result) then Exit;
    
    case iPlane of
      0: // near
      begin
        v[0].p := m_cullinfo.vecFrustum[0];
        v[1].p := m_cullinfo.vecFrustum[1];
        v[2].p := m_cullinfo.vecFrustum[2];
        v[3].p := m_cullinfo.vecFrustum[3];
        v[0].color := $80505050;
        v[1].color := $80505050;
        v[2].color := $80505050;
        v[3].color := $80505050;
      end;
      1: // far
      begin
        v[0].p := m_cullinfo.vecFrustum[4];
        v[1].p := m_cullinfo.vecFrustum[6];
        v[2].p := m_cullinfo.vecFrustum[5];
        v[3].p := m_cullinfo.vecFrustum[7];
        v[0].color := $80404040;
        v[1].color := $80404040;
        v[2].color := $80404040;
        v[3].color := $80404040;
      end;
      2: // left
      begin
        v[0].p := m_cullinfo.vecFrustum[0];
        v[1].p := m_cullinfo.vecFrustum[2];
        v[2].p := m_cullinfo.vecFrustum[4];
        v[3].p := m_cullinfo.vecFrustum[6];
        v[0].color := $80404040;
        v[1].color := $80404040;
        v[2].color := $80404040;
        v[3].color := $80404040;
      end;
      3:
      begin
        v[0].p := m_cullinfo.vecFrustum[1];
        v[1].p := m_cullinfo.vecFrustum[3];
        v[2].p := m_cullinfo.vecFrustum[5];
        v[3].p := m_cullinfo.vecFrustum[7];
        v[0].color := $80606060;
        v[1].color := $80606060;
        v[2].color := $80606060;
        v[3].color := $80606060;
      end;
      4:
      begin
        v[0].p := m_cullinfo.vecFrustum[2];
        v[1].p := m_cullinfo.vecFrustum[3];
        v[2].p := m_cullinfo.vecFrustum[6];
        v[3].p := m_cullinfo.vecFrustum[7];
        v[0].color := $80505050;
        v[1].color := $80505050;
        v[2].color := $80505050;
        v[3].color := $80505050;
      end;
      5:
      begin
        v[0].p := m_cullinfo.vecFrustum[0];
        v[1].p := m_cullinfo.vecFrustum[1];
        v[2].p := m_cullinfo.vecFrustum[4];
        v[3].p := m_cullinfo.vecFrustum[5];
        v[0].color := $80505050;
        v[1].color := $80505050;
        v[2].color := $80505050;
        v[3].color := $80505050;
      end;
    end;
    m_pPlaneVB[iPlane].Unlock;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  vecMin, vecMax: TD3DXVector3;
  pVertices: PByte;
  pCullableThing: CCullableThing;
  iThing: DWORD;
  matTrans, matScale: TD3DXMatrix;
  iPlane: Integer;
begin
  // Initialize the font's internal textures
  Result := m_pFont.InitDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;
  Result := m_pFontSmall.InitDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  // Create a teapot mesh
  Result := D3DXCreateTeapot(m_pd3dDevice, m_pMeshTeapot, nil);
  if FAILED(Result) then Exit;
  D3DXComputeNormals(m_pMeshTeapot, nil);

  // Determine bounding box of the teapot
  Result:= m_pMeshTeapot.LockVertexBuffer(D3DLOCK_READONLY, Pointer(pVertices));
  if FAILED(Result) then Exit;
  D3DXComputeBoundingBox(PD3DXVector3(pVertices), m_pMeshTeapot.GetNumVertices,
    D3DXGetFVFVertexSize(m_pMeshTeapot.GetFVF), vecMin, vecMax);
  m_pMeshTeapot.UnlockVertexBuffer;

  // Note that the m_vecBoundsLocal are identical for every CullableThing
  // because the cullable things are all the same teapot object.  In a real
  // app, of course, they would all be potentially different.
  for iThing:= 0 to  m_dwNumCullableThings - 1 do
  begin
    pCullableThing := m_CullableThingArray[iThing];
    pCullableThing.m_vecBoundsLocal[0] := D3DXVector3(vecMin.x, vecMin.y, vecMin.z); // xyz
    pCullableThing.m_vecBoundsLocal[1] := D3DXVector3(vecMax.x, vecMin.y, vecMin.z); // Xyz
    pCullableThing.m_vecBoundsLocal[2] := D3DXVector3(vecMin.x, vecMax.y, vecMin.z); // xYz
    pCullableThing.m_vecBoundsLocal[3] := D3DXVector3(vecMax.x, vecMax.y, vecMin.z); // XYz
    pCullableThing.m_vecBoundsLocal[4] := D3DXVector3(vecMin.x, vecMin.y, vecMax.z); // xyZ
    pCullableThing.m_vecBoundsLocal[5] := D3DXVector3(vecMax.x, vecMin.y, vecMax.z); // XyZ
    pCullableThing.m_vecBoundsLocal[6] := D3DXVector3(vecMin.x, vecMax.y, vecMax.z); // xYZ
    pCullableThing.m_vecBoundsLocal[7] := D3DXVector3(vecMax.x, vecMax.y, vecMax.z); // XYZ
    pCullableThing.UpdateMatrix;
  end;

  // Set up m_matBox to be able to render the bounding box properly
  D3DXMatrixTranslation(matTrans, (vecMin.x + vecMax.x) / 2,
       (vecMin.y + vecMax.y) / 2, (vecMin.z + vecMax.z) / 2 );
  D3DXMatrixScaling(matScale, vecMax.x - vecMin.x, vecMax.y - vecMin.y, vecMax.z - vecMin.z);
  // m_matBox = matScale * matTrans;
  D3DXMatrixMultiply(m_matBox, matScale, matTrans);

  Result:= D3DXCreateBox( m_pd3dDevice, 1.0, 1.0, 1.0, m_pMeshBox, nil);
  if FAILED(Result) then Exit;
  D3DXComputeNormals(m_pMeshBox, nil);

  // Create VBs for frustum planes
  for iPlane:= 0 to 5 do
  begin
    Result:= m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TPlaneVertex),
      D3DUSAGE_WRITEONLY, TPlaneVertex_FVF, D3DPOOL_MANAGED, m_pPlaneVB[iPlane], nil);
    if FAILED(Result) then Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  fAspect: Single;
  light: TD3DLight9;
  vecLightDirUnnormalized: TD3DXVector3;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);

  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,       D3DBLEND_SRCALPHA);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND,      D3DBLEND_INVSRCALPHA);

  // Set up the matrices
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(m_matProjRight, D3DX_PI/4, fAspect, 0.2, 10.0);
  D3DXMatrixPerspectiveFovLH(m_matProjLeft,  D3DX_PI/4, fAspect, 0.2, 200.0);

  // Set up a light
  {if (( m_d3dCaps.VertexProcessingCaps & D3DVTXPCAPS_DIRECTIONALLIGHTS ) ||
     !( m_dwCreateFlags & D3DCREATE_HARDWARE_VERTEXPROCESSING ) ) }
  if (m_d3dCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS <> 0)  or
     (m_dwCreateFlags and D3DCREATE_HARDWARE_VERTEXPROCESSING = 0) then
  begin
    vecLightDirUnnormalized := D3DXVector3(0.2, -0.4, 0.2);
    ZeroMemory(@light, SizeOf(TD3DLight9));
    light._Type       := D3DLIGHT_DIRECTIONAL;
    light.Diffuse.r   := 1.0;
    light.Diffuse.g   := 1.0;
    light.Diffuse.b   := 1.0;
    D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
    light.Position.x   := 0.2;
    light.Position.y   := -0.4;
    light.Position.z   := 0.2;
    light.Range        := 1000.0;
    m_pd3dDevice.SetLight(0, light);
    m_pd3dDevice.LightEnable(0, True);
  end;
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, $ff555555);

  // Set up materials
  ZeroMemory(@m_mtrlInside, SizeOf(TD3DMaterial9));
  m_mtrlInside.Diffuse.r := 0.2; m_mtrlInside.Ambient.r := 0.2; // dark green
  m_mtrlInside.Diffuse.g := 0.6; m_mtrlInside.Ambient.g := 0.6;
  m_mtrlInside.Diffuse.b := 0.2; m_mtrlInside.Ambient.b := 0.2;
  m_mtrlInside.Diffuse.a := 0.5; m_mtrlInside.Ambient.a := 0.5;
  ZeroMemory(@m_mtrlOutside, SizeOf(TD3DMaterial9));
  m_mtrlOutside.Diffuse.r := 0.6; m_mtrlOutside.Ambient.r := 0.6; // dark red
  m_mtrlOutside.Diffuse.g := 0.2; m_mtrlOutside.Ambient.g := 0.2;
  m_mtrlOutside.Diffuse.b := 0.2; m_mtrlOutside.Ambient.b := 0.2;
  m_mtrlOutside.Diffuse.a := 0.5; m_mtrlOutside.Ambient.a := 0.5;
  ZeroMemory(@m_mtrlInsideSlow, SizeOf(TD3DMaterial9));
  m_mtrlInsideSlow.Diffuse.r := 0.7; m_mtrlInsideSlow.Ambient.r := 0.7; // pastel green
  m_mtrlInsideSlow.Diffuse.g := 1.0; m_mtrlInsideSlow.Ambient.g := 1.0;
  m_mtrlInsideSlow.Diffuse.b := 0.7; m_mtrlInsideSlow.Ambient.b := 0.7;
  m_mtrlInsideSlow.Diffuse.a := 0.5; m_mtrlInsideSlow.Ambient.a := 0.5;
  ZeroMemory(@m_mtrlOutsideSlow, SizeOf(TD3DMaterial9));
  m_mtrlOutsideSlow.Diffuse.r := 1.0; m_mtrlOutsideSlow.Ambient.r := 1.0; // pastel red
  m_mtrlOutsideSlow.Diffuse.g := 0.7; m_mtrlOutsideSlow.Ambient.g := 0.7;
  m_mtrlOutsideSlow.Diffuse.b := 0.7; m_mtrlOutsideSlow.Ambient.b := 0.7;
  m_mtrlOutsideSlow.Diffuse.a := 0.5; m_mtrlOutsideSlow.Ambient.a := 0.5;
  ZeroMemory(@m_mtrlWhite, SizeOf(TD3DMaterial9));
  m_mtrlWhite.Diffuse.r := 1.0; m_mtrlWhite.Ambient.r := 1.0; // white
  m_mtrlWhite.Diffuse.g := 1.0; m_mtrlWhite.Ambient.g := 1.0;
  m_mtrlWhite.Diffuse.b := 1.0; m_mtrlWhite.Ambient.b := 1.0;
  m_mtrlWhite.Diffuse.a := 0.5; m_mtrlWhite.Ambient.a := 0.5;

  UpdateCullInfo(m_cullinfo, m_CameraRight.m_matView, m_matProjRight);
  CullObjects;
  UpdatePlaneVBs;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
var
  iPlane: Integer;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;
  SAFE_RELEASE(m_pMeshTeapot);
  SAFE_RELEASE(m_pMeshBox);

  for iPlane:= 0 to 5 do
    SAFE_RELEASE(m_pPlaneVB[iPlane]);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  // Need to support post-pixel processing (for alpha blending)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
      adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
      D3DRTYPE_SURFACE, backBufferFormat)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  x: Integer;
begin
  // Record key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    m_bKey[Char(wParam)] := True;
  end;
  // Perform commands when keys are released
  if (WM_KEYUP = uMsg) then
  begin
    m_bKey[Char(wParam)]:= False;
    case wParam of
      Ord('N'): // Set left camera to match right camera
        m_CameraLeft := m_CameraRight;
      Ord('M'): // Reset right camera to original orientation
      begin
        ZeroMemory(@m_CameraRight, SizeOf(m_CameraRight));
        UpdateCamera(m_CameraRight);
        UpdateCullInfo(m_cullinfo, m_CameraRight.m_matView, m_matProjRight);
        CullObjects;
        UpdatePlaneVBs;
      end;
      VK_F1:
        m_bShowHelp := not m_bShowHelp;
    end;
  end;
  if (WM_LBUTTONDOWN = uMsg) then
  begin
    x := LOWORD(lParam);
    if (x < Integer(m_d3dsdBackBuffer.Width div 2)) then
      m_bLeftActive := True
    else
      m_bLeftActive := False;
  end;

  Result:= inherited MsgProc(hwnd, uMsg, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: CullObjects()
// Desc: Cull each object in the CCullableThing array
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.CullObjects;
var
  pCullableThing: CCullableThing;
  iThing: DWORD;
begin
  for iThing:= 0 to m_dwNumCullableThings - 1 do
  begin
    pCullableThing := m_CullableThingArray[iThing];
    pCullableThing.m_cullstate := CullObject(m_cullinfo,
      @pCullableThing.m_vecBoundsWorld, @pCullableThing.m_planeBoundsWorld);
  end;
end;

//-----------------------------------------------------------------------------
// Name: UpdateCullInfo()
// Desc: Sets up the frustum planes, endpoints, and center for the frustum
//       defined by a given view matrix and projection matrix.  This info will
//       be used when culling each object in CullObject().
//-----------------------------------------------------------------------------
procedure UpdateCullInfo(var pCullInfo: TCullInfo; const pMatView, pMatProj: TD3DXMatrix);
// VOID UpdateCullInfo( CULLINFO* pCullInfo, D3DXMATRIX* pMatView, D3DXMATRIX* pMatProj )
var
  mat: TD3DXMatrix;
  i: Integer;
begin
  D3DXMatrixMultiply(mat, pMatView, pMatProj);
  D3DXMatrixInverse(mat, nil, mat);

  pCullInfo.vecFrustum[0] := D3DXVector3(-1.0, -1.0,  0.0); // xyz
  pCullInfo.vecFrustum[1] := D3DXVector3( 1.0, -1.0,  0.0); // Xyz
  pCullInfo.vecFrustum[2] := D3DXVector3(-1.0,  1.0,  0.0); // xYz
  pCullInfo.vecFrustum[3] := D3DXVector3( 1.0,  1.0,  0.0); // XYz
  pCullInfo.vecFrustum[4] := D3DXVector3(-1.0, -1.0,  1.0); // xyZ
  pCullInfo.vecFrustum[5] := D3DXVector3( 1.0, -1.0,  1.0); // XyZ
  pCullInfo.vecFrustum[6] := D3DXVector3(-1.0,  1.0,  1.0); // xYZ
  pCullInfo.vecFrustum[7] := D3DXVector3( 1.0,  1.0,  1.0); // XYZ

  for i:= 0 to 7 do
    D3DXVec3TransformCoord(pCullInfo.vecFrustum[i], pCullInfo.vecFrustum[i], mat);

  D3DXPlaneFromPoints(pCullInfo.planeFrustum[0], pCullInfo.vecFrustum[0],
    pCullInfo.vecFrustum[1], pCullInfo.vecFrustum[2]); // Near
  D3DXPlaneFromPoints(pCullInfo.planeFrustum[1], pCullInfo.vecFrustum[6],
    pCullInfo.vecFrustum[7], pCullInfo.vecFrustum[5]); // Far
  D3DXPlaneFromPoints(pCullInfo.planeFrustum[2], pCullInfo.vecFrustum[2],
    pCullInfo.vecFrustum[6], pCullInfo.vecFrustum[4]); // Left
  D3DXPlaneFromPoints(pCullInfo.planeFrustum[3], pCullInfo.vecFrustum[7],
    pCullInfo.vecFrustum[3], pCullInfo.vecFrustum[5]); // Right
  D3DXPlaneFromPoints(pCullInfo.planeFrustum[4], pCullInfo.vecFrustum[2],
    pCullInfo.vecFrustum[3], pCullInfo.vecFrustum[6]); // Top
  D3DXPlaneFromPoints(pCullInfo.planeFrustum[5], pCullInfo.vecFrustum[1],
    pCullInfo.vecFrustum[0], pCullInfo.vecFrustum[4]); // Bottom
end;

//-----------------------------------------------------------------------------
// Name: CullObject()
// Desc: Determine the cullstate for an object bounding box (OBB).
//       The algorithm is:
//       1) If any OBB corner pt is inside the frustum, return CS_INSIDE
//       2) Else if all OBB corner pts are outside a single frustum plane,
//          return CS_OUTSIDE
//       3) Else if any frustum edge penetrates a face of the OBB, return
//          CS_INSIDE_SLOW
//       4) Else if any OBB edge penetrates a face of the frustum, return
//          CS_INSIDE_SLOW
//       5) Else if any point in the frustum is outside any plane of the
//          OBB, return CS_OUTSIDE_SLOW
//       6) Else return CS_INSIDE_SLOW
//-----------------------------------------------------------------------------
function CullObject(const pCullInfo: TCullInfo; pVecBounds: PD3DXVector3Array; pPlaneBounds: PD3DXPlaneArray): TCullState;
var
  bOutside: array[0..7] of Byte;
  iPoint, iPlane: Integer;
  edge: array[0..11, 0..1] of TD3DXVector3;
  face: array[0..5, 0..3] of TD3DXVector3;
  pEdge, pFace: PD3DXVector3;
  iEdge, iFace: Integer;

  edge2: array[0..11, 0..1] of TD3DXVector3;
  face2: array[0..5, 0..3] of TD3DXVector3;
begin
  ZeroMemory(@bOutside, SizeOf(bOutside));

  // Check boundary vertices against all 6 frustum planes,
  // and store result (1 if outside) in a bitfield
  for iPoint:= 0 to 7 do
  begin
    for iPlane:= 0 to 5 do
    begin
      if pCullInfo.planeFrustum[iPlane].a * pVecBounds[iPoint].x +
         pCullInfo.planeFrustum[iPlane].b * pVecBounds[iPoint].y +
         pCullInfo.planeFrustum[iPlane].c * pVecBounds[iPoint].z +
         pCullInfo.planeFrustum[iPlane].d < 0 then
      begin
        bOutside[iPoint]:= bOutside[iPoint] or (1 shl iPlane);
      end;
    end;
    // If any point is inside all 6 frustum planes, it is inside
    // the frustum, so the object must be rendered.
    if bOutside[iPoint] = 0 then
    begin
      Result:= CS_INSIDE;
      Exit;
    end;
  end;

  // If all points are outside any single frustum plane, the object is
  // outside the frustum
  if bOutside[0] and bOutside[1] and bOutside[2] and bOutside[3] and
     bOutside[4] and bOutside[5] and bOutside[6] and bOutside[7] <> 0 then
  begin
    Result:= CS_OUTSIDE;
    Exit;
  end;

  // Now see if any of the frustum edges penetrate any of the faces of
  // the bounding box
  edge[ 0][0]:= pCullInfo.vecFrustum[0]; edge[ 0][1]:= pCullInfo.vecFrustum[1]; // front bottom
  edge[ 1][0]:= pCullInfo.vecFrustum[2]; edge[ 1][1]:= pCullInfo.vecFrustum[3]; // front top
  edge[ 2][0]:= pCullInfo.vecFrustum[0]; edge[ 2][1]:= pCullInfo.vecFrustum[2]; // front left
  edge[ 3][0]:= pCullInfo.vecFrustum[1]; edge[ 3][1]:= pCullInfo.vecFrustum[3]; // front right
  edge[ 4][0]:= pCullInfo.vecFrustum[4]; edge[ 4][1]:= pCullInfo.vecFrustum[5]; // back bottom
  edge[ 5][0]:= pCullInfo.vecFrustum[6]; edge[ 5][1]:= pCullInfo.vecFrustum[7]; // back top
  edge[ 6][0]:= pCullInfo.vecFrustum[4]; edge[ 6][1]:= pCullInfo.vecFrustum[6]; // back left
  edge[ 7][0]:= pCullInfo.vecFrustum[5]; edge[ 7][1]:= pCullInfo.vecFrustum[7]; // back right
  edge[ 8][0]:= pCullInfo.vecFrustum[0]; edge[ 8][1]:= pCullInfo.vecFrustum[4]; // left bottom
  edge[ 9][0]:= pCullInfo.vecFrustum[2]; edge[ 9][1]:= pCullInfo.vecFrustum[6]; // left top
  edge[10][0]:= pCullInfo.vecFrustum[1]; edge[10][1]:= pCullInfo.vecFrustum[5]; // right bottom
  edge[11][0]:= pCullInfo.vecFrustum[3]; edge[11][1]:= pCullInfo.vecFrustum[7]; // right top

  face[0][0]:= pVecBounds[0]; face[0][1]:= pVecBounds[2]; face[0][2]:= pVecBounds[3]; face[0][3]:= pVecBounds[1]; // front
  face[1][0]:= pVecBounds[4]; face[1][1]:= pVecBounds[5]; face[1][2]:= pVecBounds[7]; face[1][3]:= pVecBounds[6]; // back
  face[2][0]:= pVecBounds[0]; face[2][1]:= pVecBounds[4]; face[2][2]:= pVecBounds[6]; face[2][3]:= pVecBounds[2]; // left
  face[3][0]:= pVecBounds[1]; face[3][1]:= pVecBounds[3]; face[3][2]:= pVecBounds[7]; face[3][3]:= pVecBounds[5]; // right
  face[4][0]:= pVecBounds[2]; face[4][1]:= pVecBounds[6]; face[4][2]:= pVecBounds[7]; face[4][3]:= pVecBounds[3]; // top
  face[5][0]:= pVecBounds[0]; face[5][1]:= pVecBounds[4]; face[5][2]:= pVecBounds[5]; face[5][3]:= pVecBounds[1]; // bottom

  pEdge:= @edge[0][0];
  for iEdge:= 0 to 11 do
  begin
    pFace:= @face[0][0];
    for iFace:= 0 to 5 do
    begin
      if EdgeIntersectsFace(PD3DXVector3Array(pEdge), PD3DXVector3Array(pFace), pPlaneBounds[iFace]) then
      begin
        Result:= CS_INSIDE_SLOW;
        Exit;
      end;
      // pFace += 4;
      Inc(pFace, 4);
    end;
    // pEdge += 2;
    Inc(pEdge, 2);
  end;

  // Now see if any of the bounding box edges penetrate any of the faces of
  // the frustum
  edge2[ 0][0]:= pVecBounds[0]; edge2[ 0][1]:= pVecBounds[1]; // front bottom
  edge2[ 1][0]:= pVecBounds[2]; edge2[ 1][1]:= pVecBounds[3]; // front top
  edge2[ 2][0]:= pVecBounds[0]; edge2[ 2][1]:= pVecBounds[2]; // front left
  edge2[ 3][0]:= pVecBounds[1]; edge2[ 3][1]:= pVecBounds[3]; // front right
  edge2[ 4][0]:= pVecBounds[4]; edge2[ 4][1]:= pVecBounds[5]; // back bottom
  edge2[ 5][0]:= pVecBounds[6]; edge2[ 5][1]:= pVecBounds[7]; // back top
  edge2[ 6][0]:= pVecBounds[4]; edge2[ 6][1]:= pVecBounds[6]; // back left
  edge2[ 7][0]:= pVecBounds[5]; edge2[ 7][1]:= pVecBounds[7]; // back right
  edge2[ 8][0]:= pVecBounds[0]; edge2[ 8][1]:= pVecBounds[4]; // left bottom
  edge2[ 9][0]:= pVecBounds[2]; edge2[ 9][1]:= pVecBounds[6]; // left top
  edge2[10][0]:= pVecBounds[1]; edge2[10][1]:= pVecBounds[5]; // right bottom
  edge2[11][0]:= pVecBounds[3]; edge2[11][1]:= pVecBounds[7]; // right top

  face[0][0]:= pCullInfo.vecFrustum[0]; face[0][1]:= pCullInfo.vecFrustum[2]; face[0][2]:= pCullInfo.vecFrustum[3]; face[0][3]:= pCullInfo.vecFrustum[1]; // front
  face[1][0]:= pCullInfo.vecFrustum[4]; face[1][1]:= pCullInfo.vecFrustum[5]; face[1][2]:= pCullInfo.vecFrustum[7]; face[1][3]:= pCullInfo.vecFrustum[6]; // back
  face[2][0]:= pCullInfo.vecFrustum[0]; face[2][1]:= pCullInfo.vecFrustum[4]; face[2][2]:= pCullInfo.vecFrustum[6]; face[2][3]:= pCullInfo.vecFrustum[2]; // left
  face[3][0]:= pCullInfo.vecFrustum[1]; face[3][1]:= pCullInfo.vecFrustum[3]; face[3][2]:= pCullInfo.vecFrustum[7]; face[3][3]:= pCullInfo.vecFrustum[5]; // right
  face[4][0]:= pCullInfo.vecFrustum[2]; face[4][1]:= pCullInfo.vecFrustum[6]; face[4][2]:= pCullInfo.vecFrustum[7]; face[4][3]:= pCullInfo.vecFrustum[3]; // top
  face[5][0]:= pCullInfo.vecFrustum[0]; face[5][1]:= pCullInfo.vecFrustum[4]; face[5][2]:= pCullInfo.vecFrustum[5]; face[5][3]:= pCullInfo.vecFrustum[1]; // bottom

  pEdge:= @edge2[0][0];
  for iEdge:= 0 to 11 do
  begin
    pFace := @face2[0][0];
    for iFace:= 0 to 5 do
    begin
      if EdgeIntersectsFace(PD3DXVector3Array(pEdge), PD3DXVector3Array(pFace), pCullInfo.planeFrustum[iFace]) then
      begin
        Result:= CS_INSIDE_SLOW;
        Exit;
      end;
      Inc(pFace, 4);
    end;
    Inc(pEdge, 2);
  end;

  // Now see if frustum is contained in bounding box
  // If any frustum corner point is outside any plane of the bounding box,
  // the frustum is not contained in the bounding box, so the object
  // is outside the frustum
  for iPlane:= 0 to 5 do
  begin
    if pPlaneBounds[iPlane].a * pCullInfo.vecFrustum[0].x +
       pPlaneBounds[iPlane].b * pCullInfo.vecFrustum[0].y +
       pPlaneBounds[iPlane].c * pCullInfo.vecFrustum[0].z +
       pPlaneBounds[iPlane].d  < 0 then
    begin
      Result:= CS_OUTSIDE_SLOW;
      Exit;
    end;
  end;

  // Bounding box must contain the frustum, so render the object
  Result:= CS_INSIDE_SLOW;
end;




//-----------------------------------------------------------------------------
// Name: EdgeIntersectsFace()
// Desc: Determine if the edge bounded by the two vectors in pEdges intersects
//       the quadrilateral described by the four vectors in pFacePoints.
//       Note: pPlane could be derived from pFacePoints using
//       D3DXPlaneFromPoints, but it is precomputed in advance for greater
//       speed.
//-----------------------------------------------------------------------------
function EdgeIntersectsFace(pEdges, pFaces: PD3DXVector3Array; const pPlane: TD3DXPlane): BOOL;
{BOOL EdgeIntersectsFace( D3DXVECTOR3* pEdges, D3DXVECTOR3* pFacePoints,
                         D3DXPLANE* pPlane ) }
var
  fDist1, fDist2: Single;
  ptIntersection: TD3DXVector3;

  fAbsA, fAbsB, fAbsC: Single;
  facePoints: array[0..3] of TD3DXVector2;
  point: TD3DXVector2;
  i: Integer;

  x0, x1, x2, y0, y1, y2: Single;
  bClockwise: BOOL;
begin
  // If both edge points are on the same side of the plane, the edge does
  // not intersect the face
  fDist1 := pPlane.a * pEdges[0].x + pPlane.b * pEdges[0].y +
           pPlane.c * pEdges[0].z + pPlane.d;
  fDist2 := pPlane.a * pEdges[1].x + pPlane.b * pEdges[1].y +
           pPlane.c * pEdges[1].z + pPlane.d;
  if (fDist1 > 0) and (fDist2 > 0)  or
     (fDist1 < 0) and (fDist2 < 0) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Find point of intersection between edge and face plane (if they're
  // parallel, edge does not intersect face and D3DXPlaneIntersectLine
  // returns NULL)
  if Assigned(D3DXPlaneIntersectLine(ptIntersection, pPlane, pEdges[0], pEdges[1])) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Project onto a 2D plane to make the pt-in-poly test easier
  fAbsA := Abs(pPlane.a);
  fAbsB := Abs(pPlane.b);
  fAbsC := Abs(pPlane.c);
  if (fAbsA > fAbsB) and (fAbsA > fAbsC) then
  begin
    // Plane is mainly pointing along X axis, so use Y and Z
    for i:= 0 to 3 do
    begin
      facePoints[i].x := pFaces[i].y;
      facePoints[i].y := pFaces[i].z;
    end;
    point.x := ptIntersection.y;
    point.y := ptIntersection.z;
  end
  else if (fAbsB > fAbsA) and (fAbsB > fAbsC) then
  begin
    // Plane is mainly pointing along Y axis, so use X and Z
    for i:= 0 to 3 do
    begin
      facePoints[i].x := pFaces[i].x;
      facePoints[i].y := pFaces[i].z;
    end;
    point.x := ptIntersection.x;
    point.y := ptIntersection.z;
  end
  else
  begin
    // Plane is mainly pointing along Z axis, so use X and Y
    for i:= 0 to 3 do
    begin
      facePoints[i].x := pFaces[i].x;
      facePoints[i].y := pFaces[i].y;
    end;
    point.x := ptIntersection.x;
    point.y := ptIntersection.y;
  end;

  // If point is on the outside of any of the face edges, it is
  // outside the face.
  // We can do this by taking the determinant of the following matrix:
  // | x0 y0 1 |
  // | x1 y1 1 |
  // | x2 y2 1 |
  // where (x0,y0) and (x1,y1) are points on the face edge and (x2,y2)
  // is our test point.  If this value is positive, the test point is
  // "to the left" of the line.  To determine whether a point needs to
  // be "to the right" or "to the left" of the four lines to qualify as
  // inside the face, we need to see if the faces are specified in
  // clockwise or counter-clockwise order (it could be either, since the
  // edge could be penetrating from either side).  To determine this, we
  // do the same test to see if the third point is "to the right" or
  // "to the left" of the line formed by the first two points.
  // See http://forum.swarthmore.edu/dr.math/problems/scott5.31.96.html
  x0 := facePoints[0].x;
  y0 := facePoints[0].y;
  x1 := facePoints[1].x;
  y1 := facePoints[1].y;
  x2 := facePoints[2].x;
  y2 := facePoints[2].y;
  bClockwise:= (x1*y2 - y1*x2 - x0*y2 + y0*x2 + x0*y1 - y0*x1 < 0 );
  x2 := point.x;
  y2 := point.y;
  for i:= 0 to 3 do
  begin
    x0 := facePoints[i].x;
    y0 := facePoints[i].y;
    if i < 3 then
    begin
      x1 := facePoints[i+1].x;
      y1 := facePoints[i+1].y;
    end else
    begin
      x1 := facePoints[0].x;
      y1 := facePoints[0].y;
    end;
    if ((x1*y2 - y1*x2 - x0*y2 + y0*x2 + x0*y1 - y0*x1 > 0) = bClockwise) then
    begin
      Result:= False;
      Exit;
    end;
  end;

  // If we get here, the point is inside all four face edges,
  // so it's inside the face.
  Result:= True;
end;

end.

